//------------------------------------------------------------------------------
// name: poem-acrostic.ck
// desc: generate an acrostic poem from a given word + 
//       the words generated are synonyms or related words to the 
//       given word based on similar word embeddings
// NOTE: this needs to run from command line chuck
//       (due to ConsoleInput)
//       > chuck poem-acrostic.ck
//
// author: Terry Feng
// date: Spring 2023
//------------------------------------------------------------------------------

// instantiate
Word2Vec model;
// pre-trained model to load
//me.dir() + "utility-filter/glove-wiki-gigaword-50-tsne-2-filtered.txt" => string filepath;
me.dir() + "utility-filter/glove-wiki-gigaword-50-filtered.txt" => string filepath;

// load pre-trained model (see URLs above for download)
if( !model.load( filepath ) )
{
    <<< "cannot load model:", filepath >>>;
    me.exit();
}

50 => int K_NEAREST; // number of words to search
500.0 => float wait; // duration per word

// make a ConsoleInput
ConsoleInput in;
// tokenizer
StringTokenizer tok;
// line
string line[0];

// sound
ModalBar bar => JCRev reverb => dac;
0.01 => reverb.mix;

// ding!
9 => bar.preset;
[61,64,66,69] @=> int notes[];

spork ~ keyboard();

// spork to run a function in parellel
// some background
fun void playNote(int note)
{
    notes[note] => Std.mtof => bar.freq;
    1 => bar.noteOn;
}

// loop
while( true )
{
    // prompt
    in.prompt( "ACROSTIFY A WORD>" ) => now;

    // read
    while( in.more() )
    {
        // clear tokens array
        line.clear();
        // get the input line as one string; send to tokenizer
        tok.set( in.getLine() );
        // tokenize line into words separated by white space
        while( tok.more() )
        {
            // put into array (lowercased)
            line << tok.next().lower();
        }

        // at this point, for example,
        // "I like caTs" becomes ["i", "like", "cats"]
        // this array of strings are stored in 'line'
         
        // if non-empty
        if( line.size() )
        {
            // do something with the input
            execute( line[0] );
        }
    }
}

// do something with the text input 
fun void execute( string input )
{

    // you can either start using the input creatively
    // OR you can use it to build a command prompt, e.g.,
    if( input == "exit" || input == "quit" )
    {
        // print
        chout <= "\n\"acrostify\"\nby Terry Feng\n"; chout.flush();
        me.exit();
    }

    chout.flush();

    // Acrostify
    acrostify( input );
}

// print acrostic poem
fun void acrostify( string input )
{
    // Acrostic words
    string acrostic[input.length()]; // acrostic words to fill
    input => string copy; // copy word to check letters off
    string words[K_NEAREST];

    // Get vector of the input word
    float vec[model.dim()];
    model.getVector(input, vec);

    // Initial word search
    model.getSimilar(input, K_NEAREST, words);
    1 => int numSearches;

    // Vector of last word
    float lastWordVec[model.dim()];
    model.getVector(words[K_NEAREST-1], lastWordVec);
    float tangentVec[model.dim()];
    W2V.minus(tangentVec, vec, lastWordVec);

    while (!empty(copy))
    {
        for (0 => int i; i < input.length(); i++)
        {
            input.charAt(i) => int letter; // letter to check for

            // make sure letter isn't already filled
            if (copy.charAt(i) != ' ')
            {
                string startLetterWords[0]; // words that start with letter
                int indices[0];


                // Get all words starting with the letter
                for (0 => int j; j < words.size(); j++)
                {
                    words[j].charAt(0) => int wordLetter; // first letter of word
                    

                    if (letter == wordLetter)
                    {
                        startLetterWords << words[j]; // add word to list
                        indices << j; // add index to list
                    }
                }

                if (startLetterWords.size())
                {
                    // Pick random word from list
                    Math.random2(0, startLetterWords.size() - 1) => int index;
                    startLetterWords[index] => string currWord;

                    if (wordInList(currWord, acrostic))
                    {
                        // If word is already in acrostic, pick another word
                        continue;
                    }

                    if (currWord == input)
                    {
                        // If word is the same as the input, pick another word
                        continue;
                    }

                    currWord => acrostic[i]; // add word to acrostic
                    copy.setCharAt(i, ' '); // remove letter from copy
                    " " => words[indices[index]]; // delete word from list to avoid repeats
                }
            }
        }
        
        // Re-search based on last word
        // Traverse along tangent vector
        W2V.add(lastWordVec, tangentVec);
        model.getSimilar(lastWordVec, K_NEAREST, words);
    }

    // Print acrostic
    printAcrostic(acrostic);
}

fun void printAcrostic(string acrostic[])
{
    0.1 => reverb.mix;

    for (0 => int i; i < acrostic.size(); i++)
    {
        acrostic[i] => string curr;

        // Letter
        chout <= curr.substring(0,1).upper() + " ";
        chout.flush();
        spork ~ playNote(1);

        (wait*2/3)::ms => now;

        // Word
        chout <= curr.substring(1) + " " <= IO.newline();
        chout.flush();
        // if not last word
        if (i != acrostic.size() - 1)
        {
            spork ~ playNote(2);
        } else {
            spork ~ playNote(3);
        }
        (wait/3.0)::ms => now;
    }

    chout <= IO.newline(); chout.flush();
    
    (wait*2)::ms => now;
    0.01 => reverb.mix;
}

// check if string is empty (all spaces)
fun int empty ( string input )
{
    for (0 => int i; i < input.length(); i++)
    {
        if (input.charAt(i) != ' ')
        {
            return 0;
        }
    }

    return 1;
}

fun int wordInList(string word, string list[])
{
    for (0 => int i; i < list.size(); i++)
    {
        if (word == list[i])
        {
            return 1;
        }
    }

    return 0;
}

// word2vec helper
class W2V
{
    // result stored in x
    fun static void add( float x[], float y[] )
    {
        // smaller of the two
        x.size() < y.size() ? x.size() : y.size() => int size;
        // add into element
        for( int i; i < size; i++ ) y[i] +=> x[i];
    }
    
    // result stored in x
    fun static void minus( float z[], float x[], float y[])
    {
        // smaller of the two
        x.size() < y.size() ? x.size() : y.size() => int size;
        // add into element
        for (int i; i < size; i++)
        {
            y[i] - x[i] => z[i];
        }
    }
}

fun void keyboard()
{
    Hid hi;
    HidMsg msg;

    // which keyboard
    0 => int device;
    // get from command line
    if( me.args() ) me.arg(0) => Std.atoi => device;

    // open keyboard (get device number from command line)
    if( !hi.openKeyboard( device ) ) me.exit();

    // infinite event loop
    while( true )
    {
        // wait on event
        hi => now;

        // get one or more messages
        while( hi.recv( msg ) )
        {
            // check for action type
            if( msg.isButtonDown() )
            {
                spork ~ playNote(0);
            }
        }
    }
}