//------------------------------------------------------------------------------
// name: poem-acrostic.ck
// desc: generate an acrostic poem from a given word + 
//       the words generated are synonyms or related words to the 
//       given word based on similar word embeddings
// NOTE: this needs to run from command line chuck
//       (due to ConsoleInput)
//       > chuck poem-acrostic.ck
//
// author: Terry Feng
// date: Spring 2023
//------------------------------------------------------------------------------

// instantiate
Word2Vec model;
// pre-trained model to load
me.dir() + "utility-filter/glove-wiki-gigaword-50-tsne-2-filtered.txt" => string filepath;

// load pre-trained model (see URLs above for download)
if( !model.load( filepath ) )
{
    <<< "cannot load model:", filepath >>>;
    me.exit();
}

1000 => int K_NEAREST; // number of words to search
300.0 => float wait1; // duration per word
600.0 => float wait2; // duration per word
900.0 => float wait3; // duration per word
7 => int NUM_STANZAS; // number of stanzas

// get the nearest words
string words[K_NEAREST];
"trees" => global string noun1; "fly" => string verb1;
string noun2;

// POEM
"if _ could _ N " +
"would _ _" => string verseString;

// AUDIO
0 => int pitch;
[0,0,4,6] @=> int notes[];

fun void music()
{
    Blit b => JCRev l => dac.left;
    b => JCRev r => dac.right;
    l.mix(0.05);
    r.mix(0.05);

    l => DelayL dl => l; 0.1 => dl.gain; wait1::ms*1.5 => dl.max => dl.delay;
    r => DelayL dr => r; 0.05 => dr.gain; wait1::ms*2 => dr.max => dr.delay;

    // Constant notes
    while (true)
    {
        (37 + (12 * Math.random2(0,2)) + notes[pitch]) => Std.mtof => b.freq;
        Math.random2(1,5) => b.harmonics;
        wait1::ms => now;
    }
}


// tokenize the poem
StringTokenizer tok;
string verse[0];
tok.set( verseString );
while ( tok.more() )
{
    verse << tok.next();
}

0 => int specialCount;

// Poem Generator
spork ~ music();
for (0 => int i; i < NUM_STANZAS; i++)
{
    stanza();
    chout <= IO.newline(); chout <= IO.newline();
}
closing();

fun void closing()
{
    say("oh,"); play(2);
    say("if"); play(1);
    say("only"); play(3);
    say("trees"); play(2);
    say("could"); play(1);
    say("fly..."); play(3);
    chout <= IO.newline(); chout.flush();

    say("in"); play(2);
    say("the"); play(1);
    say("sky"); play(3);
    say("in"); play(2);
    say("the"); play(1);
    say("sky..."); play(3);

    chout <= IO.newline(); chout <= IO.newline(); chout.flush();
    chout <= "\"if trees could fly\"" <= IO.newline(); 
    chout <= "by Terry Feng" <= IO.newline();
    chout.flush();
}

// verses
fun void stanza()
{
    for (0 => int i; i < verse.size(); i++)
    {
        if (verse[i] == "N")
        {
            chout <= IO.newline(); chout.flush();
        } 
        else if (verse[i] == "..." || verse[i] == "?")
        {
            chout <= verse[i]; chout.flush();
        }
        else if (verse[i] == "_")
        {
            saySpecial();
            specialCount++;
        }
        else 
        {
            say(verse[i]);
            play(1);
        }
    }
}

fun void say(string word)
{
    chout <= word <= " "; chout.flush();
}

fun void saySpecial()
{

    if (specialCount % 4 == 0)
    {
        say(noun1);
        play(2);
    }
    else if (specialCount % 4 == 1)
    {
        say(verb1 + "...");
        play(2);
    }
    else if (specialCount % 4 == 2)
    {
        model.getSimilar(noun1, K_NEAREST, words);
        words[Math.random2(0,words.size()-1)] => noun2;
        say(noun2); // hold it in a temp variable
        play(3);
    }
    else
    {
        W2V.analogy(model, noun1, verb1, noun2, 1) @=> string word[];
        word[0] => verb1;
        say(verb1 + "?");
        // Update the temp noun
        noun2 => noun1;
        play(2);
    }
}

fun void play(int pick)
{
    pick => pitch;
    (pick * wait1)::ms => now;
}


// word2vec helper
class W2V
{
    fun static void copy( float to[], float from[] )
    {
        to.size(from.size());
        // add into element
        for( int i; i < to.size(); i++ ) from[i] => to[i];
    }

    fun static float[] dup( float from[] )
    {
        float to[from.size()];
        // add into element
        for( int i; i < to.size(); i++ ) from[i] => to[i];
        // return
        return to;
    }

    // result stored in x
    fun static void add( float x[], float y[] )
    {
        // smaller of the two
        x.size() < y.size() ? x.size() : y.size() => int size;
        // add into element
        for( int i; i < size; i++ ) y[i] +=> x[i];
    }

    fun static string[] addWords(Word2Vec w, string x, string y, int K_NEAREST)
    {
        float xVec[w.dim()];
        float yVec[w.dim()];
        w.getVector(x, xVec);
        w.getVector(y, yVec);
        add(xVec, yVec);

        string similar[K_NEAREST];
        w.getSimilar(xVec, K_NEAREST, similar);
        
        return similar;
    }
    
    // result stored in x
    fun static void minus( float x[], float y[] )
    {
        // smaller of the two
        x.size() < y.size() ? x.size() : y.size() => int size;
        // add into element
        for( int i; i < size; i++ ) y[i] -=> x[i];
    }
    
    fun static void scale( float x[], float scalar )
    {
        // scale
        for( int i; i < x.size(); i++ ) scalar *=> x[i];
    }
    
    fun static void scale( float result[], float x[], float scalar )
    {
        // scale
        for( int i; i < x.size(); i++ ) scalar * x[i] => result[i];
    }
    
    fun static string[] eval( Word2Vec @ w, string expr, int k )
    {
        // init 
        int pos;
        1.0 => float multiplier;
        string word;
        float wordVector[w.dim()];
        float exprVector[w.dim()];
        
        // compute
        while( true )
        {
            expr.find(" ") => pos;
            if( pos == -1 )
            {
                w.getVector(expr, wordVector);
                for( 0 => int i; i < w.dim(); i++ )
                {
                    wordVector[i] * multiplier +=> exprVector[i];
                }
                break;
            }
            else 
            {
                expr.substring(0, pos) => word;
                w.getVector(word, wordVector);
                for( 0 => int i; i < w.dim(); i++ )
                {
                    wordVector[i] * multiplier +=> exprVector[i];
                }
                expr.substring(pos + 1) => expr;
                if( expr.charAt(0) == '+' )
                {
                    1.0 => multiplier;
                    expr.substring(2) => expr;
                }
                else
                {
                    -1.0 => multiplier;
                    expr.substring(2) => expr;
                }
            }
        }
        
        // return
        string results[k];
        w.getSimilar( exprVector, k, results );
        
        return results;
    }
    
    // logical analogy: A to B is as C is to [what this function returns]
    fun static string[] analogy( Word2Vec @ w, string A, string B, string C, int k )
    {
        // the vector sum
        B + " - " + A + " + " + C => string v;
        return eval( w, v, k );
    }
}







