//------------------------------------------------------------------------------
// name: utility-filter.ck
// desc: input: a pre-trained word vector model
//       output: a new file filtering out certain words
//       (e.g., keeping words that contain only alphabetic
//       characters, plus any specific symbols)
//
// this program:
// 1) opens a word2vec model in the format of:
//    NUMWORDS NUMDIMS
//    word [vector]
//    word [vector]
//    ...
//    word [vector]
//    word [vector]
// 2) filters out terms (can customize how in the accept() method)
// 3) outputs a new word2vec file
//
// usage:
// > chuck utility-filter.ck:[original file]:[output file]
//
// example:
// > chuck utility-filter.ck:glove-wiki-gigaword-50-tsne-2.txt:glove-wiki-gigaword-50-tsne-2-filtered.txt
//------------------------------------------------------------------------------

if( me.args() < 2 )
{
    <<< "usage: chuck utility-filter.ck:[original file]:[output file]", "" >>>;
    me.exit();
}

// get input
me.arg(0) => string INPUT;
me.arg(1) => string OUTPUT;

// instantiate a file reader
FileIO fin;
// instantiate a file writer
FileIO fout;
// a string tokenizer
StringTokenizer tokenizer;
// a line of text
string line;
// a word token
string word;

// open a file
fin.open( INPUT, FileIO.READ );
// ensure it's ok
if( !fin.good() )
{
    cherr <= "can't open file: " <= INPUT <= " for reading..."
          <= IO.newline();
    me.exit();
}

// open a file
fout.open( OUTPUT, FileIO.WRITE );
// ensure it's ok
if( !fout.good() )
{
    cherr <= "can't open file: " <= OUTPUT <= " for writing..."
          <= IO.newline();
    me.exit();
}

int NUM_WORDS;
int NUM_DIMS;

// read and skip header line
while( fin.more() )
{
    fin.readLine() => line;
    if( line.trim() != "" )
    {
        // read header line
        tokenizer.set( line );
        // read in 
        if( tokenizer.more() )
        {
            // don't worry about number of words
            Std.atoi(tokenizer.next()) => NUM_WORDS;
            // get num dims tho
            if( tokenizer.more() )
                Std.atoi(tokenizer.next()) => NUM_DIMS;
        }
        // done with header
        break;
    }
}

// array
string lines[NUM_WORDS];
// reset
0 => NUM_WORDS;

// loop until end
while( fin.more() )
{
    // read current line
    fin.readLine() => line;
    // for each line, tokenize
    tokenizer.set( line );
    // if not empty
    if( tokenizer.more() )
    {
        // get the next word
        tokenizer.next() => word;
        // check
        if( accept( word.lower(), "_" ) && NUM_WORDS < lines.capacity() )
        {
            // save
            line => lines[NUM_WORDS];
            // increment
            NUM_WORDS++;
        }
    }
}

// output the header
fout <= NUM_WORDS <= " " <= NUM_DIMS <= IO.newline();
// loop over lines
for( int i; i < NUM_WORDS; i++ )
{
    fout <= lines[i] <= IO.newline();
}

// accept or reject word? currently we are accepting alphebetic letters + symbolsToAccept
fun int accept( string word, string symbolsToAccept )
{
    int ascii;
    for( int i; i < word.length(); i++ )
    {
        // get character's ascii code
        word.charAt(i) => ascii;
        // check accepted symbols OR if letter is alphebetic
        if( symbolsToAccept.find(ascii) < 0 && ( ascii < 97 || ascii > 122 ) )
            return false;
    }
    
    return true;
}

// done
fout.close();